@@ -33,6 +33,7 @@
 #if defined(HAVE_SYS_IOCTL_H)
 #include <sys/ioctl.h>
 #endif
+#include <signal.h>
 
 bool debugging_enabled = false;		 /* global flag to enable debugging */
 bool using_syslog = false;		 /* global flag to report to syslog */
@@ -43,7 +44,7 @@ int error_count = 0;			 /* global error counter from error() */
 /*
  * In debug mode, output the given formatted string to stderr.
  */
-void debug(const char *format, ...)
+void _debug(const char *function, const char *file, int line, const char *format, ...)
 {
 	va_list ap;
 
@@ -66,7 +67,7 @@ void debug(const char *format, ...)
 		}
 		tbuf[sizeof(tbuf) - 1] = '\0';	/* enforce termination */
 
-		(void) fprintf(stderr, "[%s] ", tbuf);
+		(void) fprintf(stderr, "[%s] (%d) %s (%s:%d): ", tbuf, getpid(), function, file, line);
 		(void) vfprintf(stderr, format, ap);	/* flawfinder: ignore */
 		(void) fprintf(stderr, "\n");
 
@@ -185,7 +186,7 @@ size_t ds_leafname_pos( /*@null@ */ char *pathname)
  *
  * Returns -1 on error, in which case nothing is put into tmpnameptr.
  */
-int ds_tmpfile( /*@null@ */ char *pathname, /*@null@ */ char **tmpnameptr)
+int ds_tmpfile( /*@null@ */ char *pathname, /*@null@ */ nullable_string * tmpnameptr)
 {
 	size_t leafpos;
 	int fd;
@@ -348,7 +349,7 @@ void setproctitle(const char *format, ...)
 	if (length > space_available - 2)
 		length = space_available - 2;
 
-	(void) snprintf(base_argv[0], space_available, "%s -- %s", PACKAGE_NAME, title);
+	(void) snprintf(base_argv[0], space_available, "%s", title);
 	if (space_available > length)
 		memset(base_argv[0] + length, 0, space_available - length);
 }
@@ -370,6 +371,17 @@ void setproctitle(const char *format, ...)
 }
 
 
+/*
+ * Comparison function that calls strcmp(), for use with tfind() /
+ * tsearch().  We use this wrapper because not all libc implementations
+ * would allow "strcmp" to be used as a function pointer.
+ */
+int compare_with_strcmp(const void *a, const void *b)
+{
+	return strcmp(a, b);
+}
+
+
 /*
  * Create the parent directories of a file, if they don't already exist. 
  * The parent directories are created with mode 755.
@@ -878,3 +890,51 @@ void showParameterDefinitions(FILE * stream, size_t columns, struct parameterDef
 		fprintf(stream, "\n");
 	}
 }
+
+/*
+ * Chain-copy a string with truncation - from string_copying(7).
+ *
+ * Copy the input string "src" into the destination string "dst", truncating
+ * the destination if it would reach "end", the end of the "dst" buffer. 
+ * The "dst" string is guaranteed to be null-terminated.
+ *
+ * Returns a pointer suitable for chaining (a new "dst" for the next call).
+ *
+ * The "dst" string must already be null-terminated before the first call.
+ */
+/*@-mustdefine@ */
+/*@-compdef@ */
+/*
+ * splint - we need to mark the return value as dependent but also need
+ * "end" to be marked "out" since it doesn't need to be defined.
+ */
+/*@dependent@ */ char *stpecpy(char *dst, /*@out @ */ char *end, const char *restrict src)
+{
+	char *p;
+
+	if (src[strlen(src)] != '\0') {	    /* flawfinder: ignore */
+		(void) raise(SIGSEGV);
+		return dst;
+	}
+	/*
+	 * flawfinder - we use strlen() here to check validity; by
+	 * definition src is supposed to have been null-terminated by the
+	 * caller.
+	 */
+
+	if (dst == end)
+		return end;
+
+	/*@-unrecog@ *//* memccpy is unknown to splint. */
+	p = memccpy(dst, src, '\0', end - dst);
+	/*@+unrecog@ */
+	if (p != NULL)
+		return p - 1;
+
+	/* truncation detected */
+	end[-1] = '\0';
+	return end;
+}
+
+/*@+mustdefine@ */
+/*@+compdef@ */
