@@ -77,7 +77,7 @@ static int find_config_section(const char *name)
  * string if necessary, returning nonzero on error (and reporting the
  * error).
  */
-static int expand_config_sequences(struct sync_set_s *cf, char **strptr, char *parameter)
+static int expand_config_sequences(struct sync_set_s *cf, /*@null@ */ nullable_string * strptr, char *parameter)
 {
 	size_t orig_string_len = 0;
 	size_t new_string_len = 0;
@@ -97,8 +97,8 @@ static int expand_config_sequences(struct sync_set_s *cf, char **strptr, char *p
 	idx = 0;
 	while (idx < orig_string_len) {
 		size_t section_length;
-		/*@dependent@ */ char *to_append;
-		size_t to_append_len;
+		/*@dependent@ */ nullable_string to_append = NULL;
+		size_t to_append_len = 0;
 
 		section_length = strcspn(&(origstr[idx]), "%");
 		if (section_length > 0) {
@@ -129,26 +129,33 @@ static int expand_config_sequences(struct sync_set_s *cf, char **strptr, char *p
 			to_append_len = strlen(cf->name);
 			break;
 		case 's':
-			to_append = cf->source;
-			to_append_len = strlen(cf->source);
+			if (NULL != cf->source) {
+				to_append = cf->source;
+				to_append_len = strlen(cf->source);
+			}
 			break;
 		case 'd':
-			to_append = strrchr(cf->destination, ':');
-			to_append_len = 0;
-			if (NULL != to_append) {
-				to_append++;
-			} else {
-				to_append = cf->destination;
+			if (NULL != cf->destination) {
+				to_append = strrchr(cf->destination, ':');
+				to_append_len = 0;
+				if (NULL != to_append) {
+					to_append++;
+				} else {
+					to_append = cf->destination;
+				}
+				if (NULL != to_append) {
+					to_append_len = strlen(to_append);
+				}
 			}
-			if (NULL != to_append)
-				to_append_len = strlen(to_append);
 			break;
 		case 'h':
-			to_append = cf->destination;
-			to_append_len = strcspn(to_append, ":");
-			if (to_append_len == strlen(to_append)) {
-				to_append = "localhost";
-				to_append_len = strlen(to_append);
+			if (NULL != cf->destination) {
+				to_append = cf->destination;
+				to_append_len = strcspn(to_append, ":");
+				if (to_append_len == strlen(to_append)) {
+					to_append = "localhost";
+					to_append_len = strlen(to_append);
+				}
 			}
 			break;
 		default:
@@ -226,6 +233,10 @@ static int validate_config_section(int idx, int defaults_idx)
 			error("%s: %s: %s", config_sections[idx].name, "destination",
 			      _("this section has not defined a destination directory"));
 			rc = 1;
+		} else if (NULL != config_sections[idx].section_list_file) {
+			error("%s: %s: %s", config_sections[idx].name, "section list file",
+			      _("only the default section may define the section list file"));
+			rc = 1;
 		}
 	}
 
@@ -249,20 +260,25 @@ debug("(cf) %s: %s: %s -> %s", config_sections[idx].name, #x, "using default", c
 		dup_default_string(partial_rsync_opts);
 		dup_default_string(log_file);
 		dup_default_string(status_file);
+		/* The section list file is only set in the default section. */
+		/* dup_default_string(section_list_file); */
 #define copy_default_ulong(x) if ((false == config_sections[idx].set.x) && (false != config_sections[defaults_idx].set.x)) { \
 config_sections[idx].x = config_sections[defaults_idx].x; \
 debug("(cf) %s: %s: %s -> %lu", config_sections[idx].name, #x, "using default", config_sections[defaults_idx].x); \
 }
 		copy_default_ulong(full_interval);
 		copy_default_ulong(full_retry);
+		copy_default_ulong(full_timeout);
 		copy_default_ulong(partial_interval);
 		copy_default_ulong(partial_retry);
+		copy_default_ulong(partial_timeout);
 		copy_default_ulong(recursion_depth);
 #define copy_default_flag(x) if ((false == config_sections[idx].set.x) && (false != config_sections[defaults_idx].set.x)) { \
 config_sections[idx].x = config_sections[defaults_idx].x; \
 debug("(cf) %s: %s: %s -> %s", config_sections[idx].name, #x, "using default", config_sections[defaults_idx].x ? "yes" : "no"); \
 }
 		copy_default_flag(ignore_vanished_files);
+		copy_default_flag(track_files);
 
 		if ((0 == config_sections[idx].exclude_count)
 		    && (0 != config_sections[defaults_idx].exclude_count)) {
@@ -292,6 +308,7 @@ debug("(cf) %s: %s: %s -> %s", config_sections[idx].name, #x, "using default", c
 	expand_sequences(partial_rsync_opts);
 	expand_sequences(log_file);
 	expand_sequences(status_file);
+	expand_sequences(section_list_file);
 
 	if (NULL != config_sections[idx].change_queue) {
 		struct stat sb;
@@ -343,6 +360,7 @@ debug("(cf) %s: %s: %s -> %s", config_sections[idx].name, #x, "using default", c
 	blank_if_none(sync_lock);
 	blank_if_none(log_file);
 	blank_if_none(status_file);
+	blank_if_none(section_list_file);
 
 	debug("(cf valid) %d %s: %s", idx, config_sections[idx].name, rc == 0 ? "OK" : "FAILED");
 
@@ -355,8 +373,9 @@ debug("(cf) %s: %s: %s -> %s", config_sections[idx].name, #x, "using default", c
  */
 static int parse_config(const char *filename, int depth)
 {
-	char linebuf[4096];
-	int lineno;
+	char *linebuf_ptr;
+	size_t linebuf_size;
+	unsigned int line_number;
 	FILE *fptr;
 	struct sync_set_s *section;
 
@@ -373,32 +392,72 @@ static int parse_config(const char *filename, int depth)
 		return 1;
 	}
 
-	lineno = 0;
+	linebuf_ptr = NULL;
+	linebuf_size = 0;
+	line_number = 0;
 	section = NULL;
-	memset(linebuf, 0, sizeof(linebuf));
 
-	while ((0 == feof(fptr)) && (0 == ferror(fptr))
-	       && (NULL != fgets(linebuf, (int) (sizeof(linebuf)) - 1, fptr))) {
+	while (0 == feof(fptr)) {
+		ssize_t line_length = 0;
 		unsigned long param_ulong;
-		char param_str[4096];
-		size_t idx, len;
+		char *param_str;
+		ssize_t idx;
 
-		lineno++;
+		errno = 0;
+		line_length = getline(&linebuf_ptr, &linebuf_size, fptr);
+		if ((line_length < 0) || (NULL == linebuf_ptr)) {
+			if (0 != errno)
+				error("%s:%u: %s", filename, line_number, strerror(errno));
+			break;
+		}
+		line_number++;
 
-		if (sscanf(linebuf, " [%999[0-9A-Za-z_.-]]", param_str) == 1) {
+		/*
+		 * The length includes the delimiter, and so it should
+		 * always be > 0.
+		 */
+		if (line_length < 1)
+			continue;
+
+		/*
+		 * Replace the delimiter with \0 to terminate the string
+		 * there.
+		 */
+		line_length--;
+		linebuf_ptr[line_length] = '\0';
+
+		param_str = NULL;
+
+		/*
+		 * Note that the sscanf() "m" modifier to %[], used below,
+		 * is from POSIX.1-2008.  It is available from glibc 2.7,
+		 * and doesn't work prior to that.  On CentOS 5, for
+		 * instance, it would have to be "%a[]" instead of "%m[]".
+		 *
+		 * TODO: detect when %m[] isn't available, at compile time.
+		 * Or don't, since CentOS 5 went EOL in 2014.
+		 */
+
+		if ((sscanf(linebuf_ptr, " [%m[0-9A-Za-z_.-]]", &param_str) == 1) && (NULL != param_str)) {
 
 			/* New section */
-			debug("(cf) %s: %d: %s: %s", filename, lineno, "section", param_str);
+			debug("(cf) %s:%u: %s: %s", filename, line_number, "section", param_str);
 
 			if (find_config_section(param_str) >= 0) {
-				error("%s: %d: %s: %s", filename, lineno, param_str,
+				error("%s:%u: %s: %s", filename, line_number, param_str,
 				      _("a section with this name was already defined"));
+				free(param_str);
+				if (NULL != linebuf_ptr)
+					free(linebuf_ptr);
 				(void) fclose(fptr);
 				return 1;
 			}
 
 			if (config_sections_count >= (MAX_CONFIG_SECTIONS - 1)) {
 				error("%s: %s: %s", filename, param_str, _("maximum number of sections reached"));
+				free(param_str);
+				if (NULL != linebuf_ptr)
+					free(linebuf_ptr);
 				(void) fclose(fptr);
 				return 1;
 			}
@@ -407,18 +466,21 @@ static int parse_config(const char *filename, int depth)
 			config_sections_count++;
 
 			memset(section, 0, sizeof(config_sections[0]));
-			section->name = xstrdup(param_str);
+			section->name = param_str;
 			section->full_interval = 86400;
 			section->full_retry = 3600;
+			section->full_timeout = 0;
 			section->partial_interval = 30;
 			section->partial_retry = 300;
+			section->partial_timeout = 0;
 			section->recursion_depth = 20;
-			section->ignore_vanished_files = 0;
+			section->ignore_vanished_files = false;
+			section->track_files = true;
 
 			continue;
 
-		} else if (sscanf(linebuf, " include = %4095[^\n]", param_str)
-			   == 1) {
+		} else if ((sscanf(linebuf_ptr, " include = %m[^\n]", &param_str)
+			    == 1) && (NULL != param_str)) {
 			wordexp_t p;
 			size_t word_idx;
 			int cwdfd;
@@ -426,7 +488,7 @@ static int parse_config(const char *filename, int depth)
 			char *ptr;
 
 			/* Include another config file */
-			debug("(cf) %s: %d: %s: %s", filename, lineno, "include", param_str);
+			debug("(cf) %s:%u: %s: %s", filename, line_number, "include", param_str);
 
 			/*
 			 * Temporarily change directory to the location of
@@ -461,21 +523,25 @@ static int parse_config(const char *filename, int depth)
 			 * "-compdestroy" is used to inhibit the warning.
 			 */
 			if (wordexp(param_str, &p, WRDE_NOCMD) != 0) {
-				error("%s: %d: %s: %s: %s", filename, lineno, "include", param_str, strerror(errno));
+				error("%s:%u: %s: %s: %s", filename, line_number, "include", param_str,
+				      strerror(errno));
 				(void) fclose(fptr);
 				if (cwdfd >= 0) {
 					if (0 != fchdir(cwdfd)) {
-						error("%s: %d: %s: %s: %s", filename, lineno, "include", param_str,
+						error("%s:%u: %s: %s: %s", filename, line_number, "include", param_str,
 						      strerror(errno));
 					}
 					(void) close(cwdfd);
 				}
+				free(param_str);
+				if (NULL != linebuf_ptr)
+					free(linebuf_ptr);
 				return 1;
 			}
 
 			if (cwdfd >= 0) {
 				if (0 != fchdir(cwdfd)) {
-					error("%s: %d: %s: %s: %s", filename, lineno, "include", param_str,
+					error("%s:%u: %s: %s: %s", filename, line_number, "include", param_str,
 					      strerror(errno));
 				}
 				(void) close(cwdfd);
@@ -496,6 +562,8 @@ static int parse_config(const char *filename, int depth)
 				}
 				if (parse_config(p.we_wordv[word_idx], depth + 1) != 0) {
 					wordfree(&p);
+					if (NULL != linebuf_ptr)
+						free(linebuf_ptr);
 					return 1;
 				}
 			}
@@ -504,18 +572,21 @@ static int parse_config(const char *filename, int depth)
 
 			/*@+compdestroy@ */
 
+			free(param_str);
+
 			continue;
 
 		} else if (NULL == section) {
 
 			/* Not in a section at the moment */
 			idx = 0;
-			len = strlen(linebuf);
-			while ((idx < len) && isspace(linebuf[idx]))
+			while ((idx < line_length) && isspace(linebuf_ptr[idx]))
 				idx++;
-			if (NULL == strchr("#\r\n", linebuf[idx])) {
-				error("%s: %d: %s", filename, lineno,
+			if (NULL == strchr("#\r\n", linebuf_ptr[idx])) {
+				error("%s:%u: %s", filename, line_number,
 				      _("a section declaration is required before any parameters"));
+				if (NULL != linebuf_ptr)
+					free(linebuf_ptr);
 				(void) fclose(fptr);
 				return 1;
 			}
@@ -528,6 +599,8 @@ static int parse_config(const char *filename, int depth)
 
 		if (NULL == section) {
 			/* There should be no way for this to be reachable. */
+			if (NULL != linebuf_ptr)
+				free(linebuf_ptr);
 			(void) fclose(fptr);
 			return 1;
 		}
@@ -536,70 +609,74 @@ static int parse_config(const char *filename, int depth)
 		 * Strip comments - a hash that's either at the start of the
 		 * line or preceded by whitespace.
 		 */
-		len = strlen(linebuf);
-		for (idx = 0; idx < len; idx++) {
-			if ('#' != linebuf[idx])
+		for (idx = 0; idx < line_length; idx++) {
+			if ('#' != linebuf_ptr[idx])
 				continue;
-			if ((idx > 0) && (!isspace(linebuf[idx - 1])))
+			if ((idx > 0) && (!isspace(linebuf_ptr[idx - 1])))
 				continue;
-			linebuf[idx] = '\0';
+			linebuf_ptr[idx] = '\0';
 			break;
 		}
 
 		/*
 		 * Strip trailing whitespace.
 		 */
-		len = strlen(linebuf);
-		if (len > 0) {
-			for (idx = len - 1; idx > 0; idx--) {
-				if (!isspace(linebuf[idx]))
+		if (line_length > 0) {
+			for (idx = line_length - 1; idx > 0; idx--) {
+				if (!isspace(linebuf_ptr[idx]))
 					break;
-				linebuf[idx] = '\0';
+				linebuf_ptr[idx] = '\0';
+				line_length--;
 			}
 		}
 
 		/*@-nullderef@ */
 		/*@-unrecog@ */
 
-#define cf_string(X, Y) if (sscanf(linebuf, " " X, param_str) == 1) { \
-debug("(cf) %s: %d: %s = [%s]", filename, lineno, #Y, param_str); \
-section->Y = xstrdup(param_str); \
+#define cf_string(X, Y) if ((sscanf(linebuf_ptr, " " X, &param_str) == 1) && (NULL != param_str)) { \
+debug("(cf) %s:%u: %s = [%s]", filename, line_number, #Y, param_str); \
+section->Y = param_str; \
 continue; \
 }
-#define cf_ulong(X, Y) if (sscanf(linebuf, X, &param_ulong) == 1) { \
-debug("(cf) %s: %d: %s = [%lu]", filename, lineno, #Y, param_ulong); \
+#define cf_ulong(X, Y) if (sscanf(linebuf_ptr, X, &param_ulong) == 1) { \
+debug("(cf) %s:%u: %s = [%lu]", filename, line_number, #Y, param_ulong); \
 section->Y = param_ulong; \
 section->set.Y = true; \
 continue; \
 }
-#define cf_flag(X, Y) if (sscanf(linebuf, " " X, param_str) == 1) { \
+#define cf_flag(X, Y) if ((sscanf(linebuf_ptr, " " X, &param_str) == 1) && (NULL != param_str)) { \
 section->Y = false; \
 if (strncasecmp("yes", param_str, 3) == 0) section->Y = true; \
 if (strncasecmp("on", param_str, 2) == 0) section->Y = true; \
-debug("(cf) %s: %d: %s = [%s]", filename, lineno, #Y, section->Y ? "yes" : "no"); \
+debug("(cf) %s:%u: %s = [%s]", filename, line_number, #Y, section->Y ? "yes" : "no"); \
 section->set.Y = true; \
+free(param_str); \
 continue; \
 }
-		cf_string("source = %4095[^\n]", source);
-		cf_string("destination = %4095[^\n]", destination);
-		cf_string("source validation command = %4095[^\n]", source_validation);
-		cf_string("destination validation command = %4095[^\n]", destination_validation);
+		cf_string("source = %m[^\n]", source);
+		cf_string("destination = %m[^\n]", destination);
+		cf_string("source validation command = %m[^\n]", source_validation);
+		cf_string("destination validation command = %m[^\n]", destination_validation);
 		cf_ulong("full sync interval = %lu", full_interval);
 		cf_ulong("full sync retry = %lu", full_retry);
+		cf_ulong("full sync timeout = %lu", full_timeout);
 		cf_ulong("partial sync interval = %lu", partial_interval);
 		cf_ulong("partial sync retry = %lu", partial_retry);
+		cf_ulong("partial sync timeout = %lu", partial_timeout);
 		cf_ulong("recursion depth = %lu", recursion_depth);
-		cf_string("full sync marker file = %4095[^\n]", full_marker);
-		cf_string("partial sync marker file = %4095[^\n]", partial_marker);
-		cf_flag("ignore vanished files = %4095[^\n]", ignore_vanished_files);
-		cf_string("change queue = %4095[^\n]", change_queue);
-		cf_string("transfer list = %4095[^\n]", transfer_list);
-		cf_string("temporary directory = %4095[^\n]", tempdir);
-		cf_string("sync lock = %4095[^\n]", sync_lock);
-		cf_string("full rsync options = %4095[^\n]", full_rsync_opts);
-		cf_string("partial rsync options = %4095[^\n]", partial_rsync_opts);
-		cf_string("log file = %4095[^\n]", log_file);
-		cf_string("status file = %4095[^\n]", status_file);
+		cf_string("full sync marker file = %m[^\n]", full_marker);
+		cf_string("partial sync marker file = %m[^\n]", partial_marker);
+		cf_flag("ignore vanished files = %m[^\n]", ignore_vanished_files);
+		cf_flag("track files = %m[^\n]", track_files);
+		cf_string("change queue = %m[^\n]", change_queue);
+		cf_string("transfer list = %m[^\n]", transfer_list);
+		cf_string("temporary directory = %m[^\n]", tempdir);
+		cf_string("sync lock = %m[^\n]", sync_lock);
+		cf_string("full rsync options = %m[^\n]", full_rsync_opts);
+		cf_string("partial rsync options = %m[^\n]", partial_rsync_opts);
+		cf_string("log file = %m[^\n]", log_file);
+		cf_string("status file = %m[^\n]", status_file);
+		cf_string("section list file = %m[^\n]", section_list_file);
 
 		/*@+nullderef@ */
 		/*@+unrecog@ */
@@ -614,15 +691,18 @@ continue; \
 		 * does not recognise strncasecmp().
 		 */
 
-		if (sscanf(linebuf, " exclude = %4095[^\n]", param_str) == 1) {
-			debug("(cf) %s: %d: %s = [%s]", filename, lineno, "exclude", param_str);
+		if ((sscanf(linebuf_ptr, " exclude = %m[^\n]", &param_str) == 1) && (NULL != param_str)) {
+			debug("(cf) %s:%u: %s = [%s]", filename, line_number, "exclude", param_str);
 			if (section->exclude_count >= (MAX_EXCLUDES - 1)) {
-				error("%s: %d: %s: %s", filename, lineno, "exclude",
+				error("%s:%u: %s: %s", filename, line_number, "exclude",
 				      _("too many exclusions specified"));
+				free(param_str);
+				if (NULL != linebuf_ptr)
+					free(linebuf_ptr);
 				(void) fclose(fptr);
 				return 1;
 			}
-			section->excludes[section->exclude_count++] = xstrdup(param_str);
+			section->excludes[section->exclude_count++] = param_str;
 			continue;
 		}
 
@@ -631,16 +711,20 @@ continue; \
 		 * an invalid directive.
 		 */
 		idx = 0;
-		len = strlen(linebuf);
-		while ((idx < len) && isspace(linebuf[idx]))
+		while ((idx < line_length) && isspace(linebuf_ptr[idx]))
 			idx++;
-		if (NULL == strchr("#\r\n", linebuf[idx])) {
-			error("%s: %d: %s", filename, lineno, _("invalid configuration directive"));
+		if (NULL == strchr("#\r\n", linebuf_ptr[idx])) {
+			error("%s:%u: %s", filename, line_number, _("invalid configuration directive"));
+			if (NULL != linebuf_ptr)
+				free(linebuf_ptr);
 			(void) fclose(fptr);
 			return 1;
 		}
 	}
 
+	if (NULL != linebuf_ptr)
+		free(linebuf_ptr);
+
 	(void) fclose(fptr);
 	return 0;
 }
@@ -673,6 +757,7 @@ config_sections[cf_idx].X = NULL; \
 		free_and_clear(partial_rsync_opts);
 		free_and_clear(log_file);
 		free_and_clear(status_file);
+		free_and_clear(section_list_file);
 		for (excl_idx = 0; excl_idx < config_sections[cf_idx].exclude_count; excl_idx++) {
 			if (NULL != config_sections[cf_idx].excludes[excl_idx]) {
 				free(config_sections[cf_idx].excludes[excl_idx]);
@@ -692,6 +777,143 @@ config_sections[cf_idx].X = NULL; \
 }
 
 
+/*
+ * Write a string to a file stream, escaping it as JSON.
+ */
+static void write_json_string(FILE * fptr, /*@null@ */ const char *string)
+{
+	if (NULL == fptr)
+		return;
+	if (NULL == string) {
+		fprintf(fptr, "null");
+		return;
+	}
+
+	fprintf(fptr, "\"");
+
+	/* TODO: rewrite this to be multibyte aware so it handles UTF-8. */
+
+	for (; string[0] != '\0'; string++) {
+		char next_char = string[0];
+		switch (next_char) {
+		case '"':
+		case '\\':
+			fprintf(fptr, "\\%c", next_char);
+			break;
+		case '\n':
+			fprintf(fptr, "\\n");
+			break;
+		case '\t':
+			fprintf(fptr, "\\t");
+			break;
+		case '\b':
+			fprintf(fptr, "\\b");
+			break;
+		case '\f':
+			fprintf(fptr, "\\f");
+			break;
+		case '\r':
+			fprintf(fptr, "\\r");
+			break;
+		default:
+			if (next_char >= ' ' && next_char < (char) 127) {
+				fprintf(fptr, "%c", next_char);
+			} else {
+				unsigned int char_val = (unsigned int) ((unsigned char *) (string))[0];
+				fprintf(fptr, "\\u%04x", char_val);
+			}
+			break;
+		}
+	}
+
+	fprintf(fptr, "\"");
+}
+
+
+/*
+ * Write details of the active sections to the section list file.
+ */
+static void write_section_list_file(char *savefile, int defaults_idx)
+{
+	char *temp_filename = NULL;
+	int tmpfd, cf_idx;
+	FILE *fptr;
+	bool first_item;
+
+	if (NULL == savefile)
+		return;
+
+	tmpfd = ds_tmpfile(savefile, &temp_filename);
+	if ((tmpfd < 0) || (NULL == temp_filename)) {
+		return;
+	}
+
+	fptr = fdopen(tmpfd, "w");
+	if (NULL == fptr) {
+		error("%s: %s", temp_filename, strerror(errno));
+		(void) close(tmpfd);
+		(void) remove(temp_filename);
+		free(temp_filename);
+		return;
+	}
+
+	fprintf(fptr, "[");
+	first_item = true;
+	for (cf_idx = 0; cf_idx < config_sections_count; cf_idx++) {
+		if (!config_sections[cf_idx].selected)
+			continue;
+		if (cf_idx == defaults_idx)
+			continue;
+
+		if (first_item) {
+			first_item = false;
+		} else {
+			fprintf(fptr, ",");
+		}
+
+		fprintf(fptr, "\n  {\n");
+
+		fprintf(fptr, "    \"%s\": ", "name");
+		write_json_string(fptr, config_sections[cf_idx].name);
+		fprintf(fptr, ",\n");
+
+		fprintf(fptr, "    \"%s\": %lu,\n", "fullSyncInterval", config_sections[cf_idx].full_interval);
+		fprintf(fptr, "    \"%s\": %lu,\n", "fullSyncRetry", config_sections[cf_idx].full_retry);
+		fprintf(fptr, "    \"%s\": %lu,\n", "fullSyncTimeout", config_sections[cf_idx].full_timeout);
+		fprintf(fptr, "    \"%s\": %lu,\n", "partialSyncInterval", config_sections[cf_idx].partial_interval);
+		fprintf(fptr, "    \"%s\": %lu,\n", "partialSyncRetry", config_sections[cf_idx].partial_retry);
+		fprintf(fptr, "    \"%s\": %lu,\n", "partialSyncTimeout", config_sections[cf_idx].partial_timeout);
+
+		fprintf(fptr, "    \"%s\": ", "statusFile");
+		write_json_string(fptr, config_sections[cf_idx].status_file);
+		fprintf(fptr, "\n");
+
+		fprintf(fptr, "  }");
+	}
+	fprintf(fptr, "\n]\n");
+
+	if (0 != fchmod(tmpfd, 0644)) {
+		error("%s: %s: %s", temp_filename, "fchmod()", strerror(errno));
+	}
+
+	if (0 != fclose(fptr)) {
+		error("%s: %s", temp_filename, strerror(errno));
+		(void) remove(temp_filename);
+		free(temp_filename);
+		return;
+	}
+
+	if (rename(temp_filename, savefile) != 0) {
+		error("%s: %s", savefile, strerror(errno));
+		(void) remove(temp_filename);
+		free(temp_filename);
+		return;
+	}
+
+	free(temp_filename);
+}
+
+
 /*
  * Output program usage information.
  */
@@ -1007,6 +1229,7 @@ int main(int argc, char **argv)
 	int rc, sel_idx, cf_idx, defaults_idx;
 	bool any_sections_chosen;
 	char *env_path;
+	bool rsync_null_delimiter;
 
 #ifdef ENABLE_NLS
 	/* Initialise language translation. */
@@ -1111,6 +1334,11 @@ int main(int argc, char **argv)
 		/*@+unrecog@ */
 	}
 
+	/*
+	 * rsync has had "--from0" since 2.6.0 (2004); assume we can use it.
+	 */
+	rsync_null_delimiter = true;
+
 	/*
 	 * Become a daemon if necessary.
 	 */
@@ -1134,6 +1362,13 @@ int main(int argc, char **argv)
 	 */
 	set_signal_handlers();
 
+	/*
+	 * Write the section list file.
+	 */
+	if (NULL != config_sections[defaults_idx].section_list_file) {
+		write_section_list_file(config_sections[defaults_idx].section_list_file, defaults_idx);
+	}
+
 	/*
 	 * Main loop: maintain a child process for each selected section.
 	 */
@@ -1155,7 +1390,7 @@ int main(int argc, char **argv)
 				/* Child - run sync for this section */
 				setproctitle("%s [%s]", common_program_name, config_sections[cf_idx].name);
 				set_signal_handlers();
-				continual_sync(&(config_sections[cf_idx]));
+				continual_sync(&(config_sections[cf_idx]), rsync_null_delimiter);
 				free_options();
 				free(common_program_name);
 				exit(EXIT_SUCCESS);
