@@ -28,6 +28,8 @@
 #include <dirent.h>
 #include <utime.h>
 #include <search.h>
+#include <signal.h>
+#include <assert.h>
 
 #define ACTION_WAITING "-"
 #define ACTION_VALIDATION_SRC "VALIDATE-SOURCE"
@@ -36,10 +38,13 @@
 #define ACTION_SYNC_FULL "SYNC-FULL"
 #define ACTION_SYNC_PARTIAL_WAIT "SYNC-PARTIAL-AWAITING-LOCK"
 #define ACTION_SYNC_PARTIAL "SYNC-PARTIAL"
+#define STRING_STATUS_WAITING "-"
+#define STRING_STATUS_OK "OK"
+#define STRING_STATUS_FAILED "FAILED"
 
 
 struct sync_status_s {
-	const char *action;
+	/*@observer@ */ const char *action;
 	pid_t watcher;
 	pid_t pid;
 	time_t next_full_sync;
@@ -48,8 +53,8 @@ struct sync_status_s {
 	time_t last_partial_sync;
 	time_t last_failed_full_sync;
 	time_t last_failed_partial_sync;
-	char *last_full_sync_status;
-	char *last_partial_sync_status;
+	/*@observer@ */ const char *last_full_sync_status;
+	/*@observer@ */ const char *last_partial_sync_status;
 	int full_sync_failures;
 	int partial_sync_failures;
 	char *workdir;
@@ -57,28 +62,22 @@ struct sync_status_s {
 	char *rsync_error_file;
 };
 
-static int run_validation(struct sync_set_s *, const char *, const char *, struct sync_status_s *, const char *);
-static void run_watcher(struct sync_set_s *);
+static int run_validation(struct sync_set_s *, /*@null@ */ const char *, const char *, struct sync_status_s *,
+			  const char *);
+static void run_watcher(struct sync_set_s *, bool);
 static void update_timestamp_file(struct sync_set_s *cf, /*@null@ */ const char *);
-static int sync_full(struct sync_set_s *, struct sync_status_s *);
-static int sync_partial(struct sync_set_s *, struct sync_status_s *);
-static void log_message(const char *, const char *, ...);
+static int sync_full(struct sync_set_s *, struct sync_status_s *, bool);
+static int sync_partial(struct sync_set_s *, struct sync_status_s *, bool);
+static void log_message( /*@null@ */ const char *, const char *, ...);
 static void recursively_delete(const char *, int);
 
 
 /*
  * Return a pointer to a static buffer describing the given epoch time as
- * YYYY-MM-DD HH:MM:SS in the local time zone, or "-" if the time is 0.
+ * "YYYY-MM-DD HH:MM:SS (@epoch)" in the local time zone, or "-" if the time
+ * is 0.
  */
-/*@-statictrans @*/
-/*
- * splint warns about transferring static storage inconsistently.  There's
- * no way to return a pointer to the static buffer without triggering a
- * warning, so the warning is inhibited.
- *
- * TODO: do this another way that doesn't use a single static buffer.
- */
-static char *dump_time(time_t t)
+/*@dependent@ */ static char *dump_time(time_t t)
 {
 	static char tbuf[256];
 	struct tm *tm;
@@ -89,14 +88,12 @@ static char *dump_time(time_t t)
 		tbuf[0] = '-';
 	} else {
 		tm = localtime(&t);
-		(void) strftime(tbuf, sizeof(tbuf) - 1, "%Y-%m-%d %H:%M:%S", tm);
+		(void) strftime(tbuf, sizeof(tbuf) - 1, "%Y-%m-%d %H:%M:%S (@%s)", tm);
 	}
 
 	return tbuf;
 }
 
-/*@+statictrans @*/
-
 
 /*
  * Update the status file, if we have one.
@@ -187,15 +184,13 @@ static void update_status_file(struct sync_set_s *cf, struct sync_status_s *st)
  * the "sync_exit_now" flag will be set by a signal handler when a SIGTERM
  * is received.
  */
-void continual_sync(struct sync_set_s *cf)
+void continual_sync(struct sync_set_s *cf, bool rsync_null_delimiter)
 {
-	char workdir[4096];
 	struct sync_status_s status;
 	struct stat sb;
 	FILE *fptr;
 
-	memset(workdir, 0, sizeof(workdir));
-
+	memset(&status, 0, sizeof(status));
 	status.action = ACTION_WAITING;
 	status.watcher = 0;
 	status.pid = (pid_t) getpid();
@@ -205,78 +200,79 @@ void continual_sync(struct sync_set_s *cf)
 	status.last_partial_sync = 0;
 	status.last_failed_full_sync = 0;
 	status.last_failed_partial_sync = 0;
-	status.last_full_sync_status = "-";
-	status.last_partial_sync_status = "-";
+	status.last_full_sync_status = STRING_STATUS_WAITING;
+	status.last_partial_sync_status = STRING_STATUS_WAITING;
 	status.full_sync_failures = 0;
 	status.partial_sync_failures = 0;
-	status.workdir = workdir;
-	status.excludes_file = NULL;
+	status.workdir = NULL;
 	status.rsync_error_file = NULL;
+	status.excludes_file = NULL;
 
 	/*
 	 * Create a temporary working directory.
 	 */
-	(void) snprintf(workdir, sizeof(workdir) - 1, "%s/%s", NULL == cf->tempdir ? "/tmp" : cf->tempdir,
-			"syncXXXXXX");
 	/*@-unrecog@ */
-	/* splint doesn't know about mkdtemp() */
-	if (mkdtemp(workdir) == NULL) {
-		error("%s: %s: %s", "mkdtemp", workdir, strerror(errno));
-		return;
+	/* splint doesn't know about asprintf() */
+	if (asprintf(&(status.workdir), "%s/%s", NULL == cf->tempdir ? "/tmp" : cf->tempdir, "syncXXXXXX")
+	    < 0) {
+		error("%s: %s", "asprintf", strerror(errno));
+		status.workdir = NULL;
+		goto continual_sync_final_cleanup;
 	}
 	/*@+unrecog@ */
-	debug("%s: %s", "temporary working directory", workdir);
+	assert(NULL != status.workdir);
+	if (NULL == status.workdir) {
+		die("%s: %s", "asprintf", "NULL");
+		goto continual_sync_final_cleanup;
+	}
+	/*@-unrecog@ */
+	/* splint doesn't know about mkdtemp() */
+	if (mkdtemp(status.workdir) == NULL) {
+		error("%s: %s: %s", "mkdtemp", status.workdir, strerror(errno));
+		goto continual_sync_final_cleanup;
+	}
+	/*@+unrecog@ */
+	debug("%s: %s", "temporary working directory", status.workdir);
 
 	/*
 	 * Create the file that rsync will write stderr to.
 	 */
-	/*@-unrecog@ */
-	/* splint doesn't know about asprintf() */
-	status.rsync_error_file = NULL;
-	if (asprintf(&(status.rsync_error_file), "%s/%s", workdir, "rsync-stderr") < 0) {
+	if (asprintf(&(status.rsync_error_file), "%s/%s", status.workdir, "rsync-stderr") < 0) {
 		error("%s: %s", "asprintf", strerror(errno));
-		(void) rmdir(workdir);
-		return;
+		status.rsync_error_file = NULL;
+		goto continual_sync_final_cleanup;
 	}
-	/*@+unrecog@ */
+	assert(NULL != status.rsync_error_file);
 	if (NULL == status.rsync_error_file) {
 		die("%s: %s", "asprintf", "NULL");
-		(void) rmdir(workdir);
-		return;
+		goto continual_sync_final_cleanup;
 	}
 
 	/*
 	 * Create the file that rsync will use with --excludes-from.
 	 */
-	status.excludes_file = NULL;
-	/*@-unrecog@ */
-	/* splint doesn't know about asprintf() */
-	if (asprintf(&(status.excludes_file), "%s/%s", workdir, "excludes")
+	if (asprintf(&(status.excludes_file), "%s/%s", status.workdir, "excludes")
 	    < 0) {
 		error("%s: %s", "asprintf", strerror(errno));
-		free(status.rsync_error_file);
-		(void) rmdir(workdir);
-		return;
+		status.excludes_file = NULL;
+		goto continual_sync_final_cleanup;
 	}
-	/*@+unrecog@ */
+	assert(NULL != status.excludes_file);
 	if (NULL == status.excludes_file) {
 		die("%s: %s", "asprintf", "NULL");
-		free(status.rsync_error_file);
-		(void) rmdir(workdir);
-		return;
+		goto continual_sync_final_cleanup;
 	}
+
 	fptr = fopen(status.excludes_file, "w");
 	if (NULL == fptr) {
 		error("%s: %s: %s", status.excludes_file, "fopen", strerror(errno));
-		free(status.rsync_error_file);
-		free(status.excludes_file);
-		recursively_delete(workdir, 0);
-		return;
+		goto continual_sync_final_cleanup;
 	}
 	if (cf->exclude_count > 0) {
 		int eidx;
 		for (eidx = 0; eidx < cf->exclude_count; eidx++) {
-			fprintf(fptr, "%s\n", cf->excludes[eidx]);
+			/* Use \0 rather than \n if rsync is using null delimiters (#11). */
+			fprintf(fptr, "%s%c", cf->excludes[eidx], rsync_null_delimiter ? '\0' : '\n');
 		}
 	} else {
 		fprintf(fptr, "*.tmp\n*~\n");
@@ -290,38 +286,40 @@ void continual_sync(struct sync_set_s *cf)
 	 * temporary working directory.
 	 */
 	if (NULL == cf->transfer_list) {
-		if (asprintf(&(cf->transfer_list), "%s/%s", workdir, "transfer") < 0) {
+		if (asprintf(&(cf->transfer_list), "%s/%s", status.workdir, "transfer") < 0) {
 			error("%s: %s", "asprintf", strerror(errno));
-			free(status.rsync_error_file);
-			free(status.excludes_file);
 			cf->transfer_list = NULL;
-			recursively_delete(workdir, 0);
-			return;
+			goto continual_sync_final_cleanup;
 		}
 		debug("%s: %s", "automatically set transfer list", cf->transfer_list);
 	}
+	assert(NULL != cf->transfer_list);
+	if (NULL == cf->transfer_list) {
+		die("%s: %s", "asprintf", "NULL");
+		goto continual_sync_final_cleanup;
+	}
 
 	/*
 	 * If no change queue directory was defined, create one under the
 	 * temporary working directory.
 	 */
 	if (NULL == cf->change_queue) {
-		if (asprintf(&(cf->change_queue), "%s/%s", workdir, "changes") < 0) {
+		if (asprintf(&(cf->change_queue), "%s/%s", status.workdir, "changes") < 0) {
 			error("%s: %s", "asprintf", strerror(errno));
-			free(status.rsync_error_file);
-			free(status.excludes_file);
 			cf->change_queue = NULL;
-			recursively_delete(workdir, 0);
-			return;
+			goto continual_sync_final_cleanup;
 		}
+		assert(NULL != cf->change_queue);
+		if (NULL == cf->change_queue) {
+			die("%s: %s", "asprintf", "NULL");
+			goto continual_sync_final_cleanup;
+		}
+
 		if (mkdir(cf->change_queue, 0700) != 0) {
 			error("%s: %s: %s", cf->change_queue, "mkdir", strerror(errno));
-			free(status.rsync_error_file);
-			free(status.excludes_file);
 			free(cf->change_queue);
 			cf->change_queue = NULL;
-			recursively_delete(workdir, 0);
-			return;
+			goto continual_sync_final_cleanup;
 		}
 		debug("%s: %s", "automatically set change queue", cf->change_queue);
 	}
@@ -329,10 +327,14 @@ void continual_sync(struct sync_set_s *cf)
 	log_message(cf->log_file, "[%s] %s", cf->name, _("sync process started"));
 
 	/*
-	 * If we're using a full sync marker file, use its timestamp to
-	 * determine the next full sync time.
+	 * If we're using a full sync marker file, use its timestamp to set
+	 * our idea of the last full sync time, and to determine the next
+	 * full sync time.
 	 */
 	if ((NULL != cf->full_marker) && (stat(cf->full_marker, &sb) == 0)) {
+		status.last_full_sync = (time_t) (sb.st_mtime);
+		log_message(cf->log_file, "[%s] %s: %s", cf->name,
+			    _("set last full sync time from marker file"), dump_time(status.last_full_sync));
 		status.next_full_sync = (time_t) (sb.st_mtime + cf->full_interval);
 		log_message(cf->log_file, "[%s] %s: %s", cf->name,
 			    _("derived next full sync time from marker file"), dump_time(status.next_full_sync));
@@ -340,10 +342,14 @@ void continual_sync(struct sync_set_s *cf)
 
 	/*
 	 * If we're using a partial sync marker file, use its timestamp to
-	 * determine the next partial sync time.
+	 * set our idea of the last partial sync time, and to determine the
+	 * next partial sync time.
 	 */
 	if ((NULL != cf->partial_marker)
 	    && (stat(cf->partial_marker, &sb) == 0)) {
+		status.last_partial_sync = (time_t) (sb.st_mtime);
+		log_message(cf->log_file, "[%s] %s: %s", cf->name,
+			    _("set last partial sync time from marker file"), dump_time(status.last_partial_sync));
 		status.next_partial_sync = (time_t) (sb.st_mtime + cf->partial_interval);
 		log_message(cf->log_file, "[%s] %s: %s", cf->name,
 			    _("derived next partial sync time from marker file"), dump_time(status.next_partial_sync));
@@ -380,15 +386,21 @@ void continual_sync(struct sync_set_s *cf)
 				child = (pid_t) fork();
 				if (0 == child) {
 					/* Child - run watcher */
-					run_watcher(cf);
+					run_watcher(cf, rsync_null_delimiter);
 					/*
 					 * We return here instead of exiting
 					 * so that the main sync program can
 					 * do its usual cleanup, freeing up
 					 * memory etc.
+					 *
+					 * NB we don't jump to
+					 * continual_sync_final_cleanup
+					 * because that would also delete
+					 * the working directory.
 					 */
 					free(status.rsync_error_file);
 					free(status.excludes_file);
+					free(status.workdir);
 					return;
 				} else if (child < 0) {
 					/* Error - output a warning */
@@ -426,7 +438,7 @@ void continual_sync(struct sync_set_s *cf)
 				 * Validation succeeded - attempt to run
 				 * sync
 				 */
-				sync_rc = sync_full(cf, &status);
+				sync_rc = sync_full(cf, &status, rsync_null_delimiter);
 				if (0 == sync_rc) {
 					/* sync succeeded */
 					/*
@@ -439,7 +451,7 @@ void continual_sync(struct sync_set_s *cf)
 					status.next_full_sync = time(NULL) + cf->full_retry;
 					status.last_failed_full_sync = time(NULL);
 					status.full_sync_failures++;
-					status.last_full_sync_status = "FAILED";
+					status.last_full_sync_status = STRING_STATUS_FAILED;
 				}
 			}
 			/*
@@ -473,7 +485,7 @@ void continual_sync(struct sync_set_s *cf)
 				 * Validation succeeded - attempt to run
 				 * sync
 				 */
-				sync_rc = sync_partial(cf, &status);
+				sync_rc = sync_partial(cf, &status, rsync_null_delimiter);
 				if (0 == sync_rc) {
 					/* sync succeeded OR not run */
 					status.next_partial_sync = time(NULL) + cf->partial_interval;
@@ -487,7 +499,7 @@ void continual_sync(struct sync_set_s *cf)
 					status.next_partial_sync = time(NULL) + cf->partial_retry;
 					status.last_failed_partial_sync = time(NULL);
 					status.partial_sync_failures++;
-					status.last_partial_sync_status = "FAILED";
+					status.last_partial_sync_status = STRING_STATUS_FAILED;
 				}
 			}
 			/*
@@ -518,14 +530,14 @@ void continual_sync(struct sync_set_s *cf)
 			 * avoid too many stat calls.
 			 */
 			check_workdir = false;
-			if (stat(workdir, &sb) != 0) {
+			if (stat(status.workdir, &sb) != 0) {
 				log_message(cf->log_file, "[%s] %s",
 					    cf->name, _("exiting immediately - the working directory has disappeared"));
-				sync_exit_now = 1;
+				sync_exit_now = true;
 			} else if (stat(status.excludes_file, &sb) != 0) {
 				log_message(cf->log_file, "[%s] %s",
 					    cf->name, _("exiting immediately - the exclusions file has disappeared"));
-				sync_exit_now = 1;
+				sync_exit_now = true;
 			}
 		}
 
@@ -533,34 +545,52 @@ void continual_sync(struct sync_set_s *cf)
 			(void) usleep(100000);
 	}
 
-	/*
-	 * Free the rsync_error_file string we made.
-	 */
-	free(status.rsync_error_file);
-
-	/*
-	 * Free the excludes_file string we made.
-	 */
-	free(status.excludes_file);
-
 	/*
 	 * Kill our watcher process, if we have one.
 	 */
-	if (0 != status.watcher)
-		kill(status.watcher, SIGTERM);
-
-	/*
-	 * Remove our temporary working directory.
-	 */
-	recursively_delete(workdir, 0);
+	if (0 != status.watcher) {
+		if (0 != kill(status.watcher, SIGTERM)) {
+			log_message(cf->log_file, "[%s] %s: %d: %s", cf->name, _("failed to terminate watcher process"),
+				    status.watcher, strerror(errno));
+		}
+	}
 
 	/*
 	 * Remove our status file, if we have one.
 	 */
-	if (NULL != cf->status_file)
-		remove(cf->status_file);
+	if (NULL != cf->status_file) {
+		if (0 != remove(cf->status_file)) {
+			log_message(cf->log_file, "[%s] %s: %s: %s", cf->name, _("failed to remove status file"),
+				    cf->status_file, strerror(errno));
+		}
+	}
 
 	log_message(cf->log_file, "[%s] %s", cf->name, _("sync process ended"));
+
+	/*
+	 * Below is where we jump to if any of the initial setup, before the
+	 * main loop, failed.
+	 */
+      continual_sync_final_cleanup:
+	/*
+	 * Free the rsync_error_file string we made.
+	 */
+	if (NULL != status.rsync_error_file)
+		free(status.rsync_error_file);
+
+	/*
+	 * Free the excludes_file string we made.
+	 */
+	if (NULL != status.excludes_file)
+		free(status.excludes_file);
+
+	/*
+	 * Remove our temporary working directory and free its string.
+	 */
+	if (NULL != status.workdir) {
+		recursively_delete(status.workdir, 0);
+		free(status.workdir);
+	}
 }
 
 
@@ -571,7 +601,7 @@ void continual_sync(struct sync_set_s *cf)
  * Before the command starts, st->action is set to "action" and the
  * status file is updated.
  */
-static int run_validation(struct sync_set_s *cf, const char *command,
+static int run_validation(struct sync_set_s *cf, /*@null@ */ const char *command,
 			  const char *name, struct sync_status_s *st, const char *action)
 {
 	int ret, exit_status;
@@ -589,7 +619,7 @@ static int run_validation(struct sync_set_s *cf, const char *command,
 	if (WIFSIGNALED(ret)) {
 		log_message(cf->log_file, "[%s] %s: %s: %d",
 			    cf->name, name, _("the validation command received a signal"), WTERMSIG(ret));
-		sync_exit_now = 1;
+		sync_exit_now = true;
 		return 1;
 	}
 
@@ -608,11 +638,16 @@ static int run_validation(struct sync_set_s *cf, const char *command,
 /*
  * Run the watcher on the source directory.
  */
-static void run_watcher(struct sync_set_s *cf)
+static void run_watcher(struct sync_set_s *cf, bool rsync_null_delimiter)
 {
 	struct watch_dir_params_s params;
 	size_t exclude_idx;
 
+	if (NULL == cf->source)
+		return;
+	if (NULL == cf->change_queue)
+		return;
+
 	setproctitle("%s %s [%s]", common_program_name, _("watcher"), cf->name);
 
 	memset(&params, 0, sizeof(params));
@@ -632,11 +667,11 @@ static void run_watcher(struct sync_set_s *cf)
 		params.excludes[exclude_idx] = cf->excludes[exclude_idx];
 	}
 	params.exclude_count = cf->exclude_count;
-	params.null_delimiter = false;
+	params.null_delimiter = rsync_null_delimiter;
 	params.only_list_files = false;
 	params.only_list_directories = false;
+	params.no_file_tracking = cf->track_files ? false : true;
 
-	/* TODO: maybe produce some sort of warning if this returns an error */
 	(void) watch_dir(&params);
 }
 
@@ -644,10 +679,10 @@ static void run_watcher(struct sync_set_s *cf)
 /*
  * Run rsync with the given parameters, returning the exit status.
  */
-static int run_rsync(const char *log_file, const char *section,
-		     const char *source, const char *destination,
-		     const char *excludes_file, const char *options,
-		     const char *transfer_list, bool ignore_vanished_files, const char *rsync_error_file)
+static int run_rsync(const char *log_file, const char *section, const char *source, const char *destination,
+		     /*@null@ */ const char *excludes_file, const char *options,
+		     /*@null@ */ const char *transfer_list, bool ignore_vanished_files, bool rsync_null_delimiter,
+		     const char *rsync_error_file, unsigned long timeout)
 {
 	char **rsync_argv;
 	int rsync_argc, optidx;
@@ -655,7 +690,9 @@ static int run_rsync(const char *log_file, const char *section,
 	int rc = -1;
 	pid_t rsync_pid;
 	struct stat sb;
+	time_t expiry_time = 0;
 
+	memset(&p, 0, sizeof(p));
 	if (wordexp(options, &p, WRDE_NOCMD) != 0) {
 		error("%s: [%s]: %s", "wordexp", options, strerror(errno));
 		return -1;
@@ -667,9 +704,11 @@ static int run_rsync(const char *log_file, const char *section,
 		rsync_argc += 2;	    /* --files-from */
 	if (NULL != excludes_file)
 		rsync_argc += 2;	    /* --exclude-from */
+	if (rsync_null_delimiter)
+		rsync_argc++;		    /* --from0 */
 	rsync_argc += 2;		    /* source, dest */
 
-	rsync_argv = calloc(rsync_argc + 1, sizeof(char *));
+	rsync_argv = calloc((size_t) (rsync_argc + 1), sizeof(char *));
 	if (NULL == rsync_argv) {
 		error("%s: %s", "calloc", strerror(errno));
 		return -1;
@@ -687,11 +726,17 @@ static int run_rsync(const char *log_file, const char *section,
 		rsync_argv[optidx++] = "--exclude-from";
 		rsync_argv[optidx++] = (char *) excludes_file;
 	}
+	if (rsync_null_delimiter) {
+		rsync_argv[optidx++] = "--from0";
+	}
 	rsync_argv[optidx++] = (char *) source;
 	rsync_argv[optidx++] = (char *) destination;
 
 	remove(rsync_error_file);
 
+	if (timeout > 0)
+		expiry_time = time(NULL) + timeout;
+
 	rsync_pid = fork();
 
 	if (0 == rsync_pid) {
@@ -702,7 +747,7 @@ static int run_rsync(const char *log_file, const char *section,
 			(void) dup2(fd, 2);
 			(void) close(fd);
 		}
-		execvp("rsync", rsync_argv);
+		(void) execvp("rsync", rsync_argv);
 		exit(EXIT_FAILURE);
 	} else if (rsync_pid < 0) {
 		/* Error - output a warning */
@@ -714,18 +759,35 @@ static int run_rsync(const char *log_file, const char *section,
 		while ((!sync_exit_now) && (0 != rsync_pid)) {
 			int wait_status;
 			pid_t waited_for;
+
+			if (timeout > 0 && time(NULL) >= expiry_time) {
+				debug("(rsync) %s", "timeout reached - sending SIGKILL");
+				(void) kill(rsync_pid, SIGKILL);
+			}
+
+			if (timeout > 0)
+				(void) alarm(1);
+
 			waited_for = (pid_t) waitpid(rsync_pid, &wait_status, 0);
+
 			if (waited_for < 0) {
-				if ((errno == EINTR) || (errno == EAGAIN))
+				if ((errno == EINTR) || (errno == EAGAIN)) {
+					if (timeout > 0)
+						(void) alarm(0);
 					continue;
+				}
 				log_message(log_file, "[%s] %s: %s: %s",
 					    section, _("failed to wait for rsync"), "waitpid", strerror(errno));
 				rc = -1;
+				if (timeout > 0)
+					(void) alarm(0);
 				break;
 			} else {
 				rsync_pid = 0;
 				rc = WEXITSTATUS(wait_status);
 				debug("(rsync) %s: %d", "process ended, exit status", rc);
+				if (timeout > 0)
+					(void) alarm(0);
 			}
 		}
 		if (0 != rsync_pid) {
@@ -738,7 +800,8 @@ static int run_rsync(const char *log_file, const char *section,
 	wordfree(&p);
 
 	if ((stat(rsync_error_file, &sb) == 0) && (sb.st_size > 0)) {
-		char linebuf[1024];
+		char *linebuf_ptr = NULL;
+		size_t linebuf_size = 0;
 		FILE *err_fptr;
 
 		err_fptr = fopen(rsync_error_file, "r");
@@ -747,18 +810,29 @@ static int run_rsync(const char *log_file, const char *section,
 			return rc;
 		}
 
-		while ((!feof(err_fptr))
-		       && (NULL != fgets(linebuf, sizeof(linebuf) - 1, err_fptr))) {
-			char *nlptr;
+		while (0 == feof(err_fptr)) {
+			ssize_t line_length = 0;
 
-			nlptr = strrchr(linebuf, '\n');
-			if (NULL != nlptr)
-				nlptr[0] = '\0';
+			/* Like getdelim() in collate_transfer_list(). */
+			errno = 0;
+			line_length = getline(&linebuf_ptr, &linebuf_size, err_fptr);
+			if ((line_length < 0) || (NULL == linebuf_ptr)) {
+				if (0 != errno)
+					error("%s: %s", rsync_error_file, strerror(errno));
+				break;
+			}
+			if (line_length < 1)
+				continue;
+			line_length--;
+			linebuf_ptr[line_length] = '\0';
 
-			log_message(log_file, "[%s] %s: %s", section, "rsync", linebuf);
+			log_message(log_file, "[%s] %s: %s", section, "rsync", linebuf_ptr);
 		}
 
-		fclose(err_fptr);
+		if (NULL != linebuf_ptr)
+			free(linebuf_ptr);
+
+		(void) fclose(err_fptr);
 
 		log_message(log_file, "[%s] %s: %d", section, "rsync failed with exit status", rc);
 	}
@@ -783,7 +857,7 @@ static void update_timestamp_file(struct sync_set_s *cf, /*@null@ */ const char
 		return;
 	fptr = fopen(file, "a");
 	if (NULL != fptr) {
-		fclose(fptr);
+		(void) fclose(fptr);
 	} else {
 		log_message(cf->log_file, "[%s] %s: %s", cf->name, file, strerror(errno));
 	}
@@ -803,7 +877,7 @@ static void update_timestamp_file(struct sync_set_s *cf, /*@null@ */ const char
  *
  * The status file is updated before the sync begins.
  */
-static int sync_full(struct sync_set_s *cf, struct sync_status_s *st)
+static int sync_full(struct sync_set_s *cf, struct sync_status_s *st, bool rsync_null_delimiter)
 {
 	int lockfd = -1;
 	int rc = 0;
@@ -836,7 +910,8 @@ static int sync_full(struct sync_set_s *cf, struct sync_status_s *st)
 		       st->excludes_file,
 		       NULL ==
 		       cf->full_rsync_opts ? "--delete -axH" :
-		       cf->full_rsync_opts, NULL, cf->ignore_vanished_files, st->rsync_error_file);
+		       cf->full_rsync_opts, NULL, cf->ignore_vanished_files, rsync_null_delimiter, st->rsync_error_file,
+		       cf->full_timeout);
 
 	log_message(cf->log_file, "[%s] %s: %s: %s", cf->name,
 		    _("full sync"), _("sync ended"), rc == 0 ? _("OK") : _("FAILED"));
@@ -852,37 +927,33 @@ static int sync_full(struct sync_set_s *cf, struct sync_status_s *st)
 		update_timestamp_file(cf, cf->full_marker);
 		st->last_full_sync = time(NULL);
 		st->full_sync_failures = 0;
-		st->last_full_sync_status = "OK";
+		st->last_full_sync_status = STRING_STATUS_OK;
 	}
 
 	return rc;
 }
 
 
-/*
- * Comparison function that calls strcmp(), for use with tfind() /
- * tsearch().  We use this wrapper because not all libc implementations
- * would allow "strcmp" to be used as a function pointer.
- */
-int compare_with_strcmp(const void *a, const void *b)
-{
-	return strcmp(a, b);
-}
-
-
 /*
  * Collate a transfer list from the change queue: remove the change queue
  * entries, appending those that refer to items that still exist to the
  * transfer list.
  */
-static void collate_transfer_list(struct sync_set_s *cf)
+static void collate_transfer_list(struct sync_set_s *cf, bool rsync_null_delimiter)
 {
 	struct dirent **namelist;
 	int namelist_length, idx;
-	char path[4096] = { 0, };
 	FILE *list_fptr;
 	FILE *changefile_fptr;
 	void *tree_root = NULL;
+	char line_delimiter;
+
+	line_delimiter = rsync_null_delimiter ? '\0' : '\n';
+
+	if (NULL == cf->transfer_list)
+		return;
+	if (NULL == cf->change_queue)
+		return;
 
 	list_fptr = fopen(cf->transfer_list, "a");
 	if (NULL == list_fptr) {
@@ -893,64 +964,105 @@ static void collate_transfer_list(struct sync_set_s *cf)
 	namelist_length = scandir(cf->change_queue, &namelist, NULL, alphasort);
 	if (0 > namelist_length) {
 		error("%s: %s: %s", "scandir", cf->change_queue, strerror(errno));
-		fclose(list_fptr);
+		(void) fclose(list_fptr);
 		return;
 	}
 
 	for (idx = 0; idx < namelist_length; idx++) {
+		char *path = NULL;
 		struct stat sb;
-		char linebuf[4096] = { 0, };
+		char *linebuf_ptr;
+		size_t linebuf_size;
+		unsigned int line_number;
 
 		if ('.' == namelist[idx]->d_name[0])
 			continue;
 
-		snprintf(path, sizeof(path) - 1, "%s/%s", cf->change_queue, namelist[idx]->d_name);
+		if (asprintf(&path, "%s/%s", cf->change_queue, namelist[idx]->d_name) < 0) {
+			error("%s: %s", "asprintf", strerror(errno));
+			continue;
+		}
+		assert(NULL != path);
+		if (NULL == path)
+			continue;
 
-		if (lstat(path, &sb) != 0)
+		memset(&sb, 0, sizeof(sb));
+		if (lstat(path, &sb) != 0) {
+			free(path);
 			continue;
-		if (!S_ISREG(sb.st_mode))
+		}
+		if (!S_ISREG(sb.st_mode)) {
+			free(path);
 			continue;
+		}
 
 		changefile_fptr = fopen(path, "r");
 		if (NULL == changefile_fptr) {
 			debug("%s: %s", path, strerror(errno));
 			remove(path);
+			free(path);
 			continue;
 		}
 
-		while ((!feof(changefile_fptr))
-		       && (NULL != fgets(linebuf, sizeof(linebuf) - 1, changefile_fptr))) {
-			char *nlptr;
+		linebuf_ptr = NULL;
+		linebuf_size = 0;
+		line_number = 0;
+
+		while (0 == feof(changefile_fptr)) {
+			ssize_t line_length = 0;
 			char *changedpath;
 
-			nlptr = strrchr(linebuf, '\n');
-			if (NULL != nlptr)
-				nlptr[0] = '\0';
+			errno = 0;
+			line_length = getdelim(&linebuf_ptr, &linebuf_size, line_delimiter, changefile_fptr);
+			if ((line_length < 0) || (NULL == linebuf_ptr)) {
+				if (0 != errno)
+					error("%s:%u: %s", path, line_number, strerror(errno));
+				break;
+			}
+			line_number++;
 
 			/*
-			 * Use a binary tree to keep track of lines we've
+			 * The length includes the delimiter, and so it
+			 * should always be > 0.
+			 */
+			if (line_length < 1)
+				continue;
+
+			/*
+			 * Replace the delimiter with \0 to terminate the
+			 * string there.
+			 */
+			line_length--;
+			linebuf_ptr[line_length] = '\0';
+
+			/*
+			 * Use a binary tree to keep track of entries we've
 			 * seen before, so we can strip duplicates.
 			 */
 			if ((NULL != tree_root)
-			    && (NULL != tfind(linebuf, &tree_root, compare_with_strcmp))) {
-				debug("%s: %s", "skipping duplicate change line", linebuf);
+			    && (NULL != tfind(linebuf_ptr, &tree_root, compare_with_strcmp))) {
+				debug("%s: %s", "skipping duplicate change item", linebuf_ptr);
 				continue;
 			}
-			tsearch(xstrdup(linebuf), &tree_root, compare_with_strcmp);
+			tsearch(xstrdup(linebuf_ptr), &tree_root, compare_with_strcmp);
 
-			if (asprintf(&changedpath, "%s/%s", cf->source, linebuf) < 0) {
+			if (asprintf(&changedpath, "%s/%s", cf->source, linebuf_ptr) < 0) {
 				error("%s: %s", "asprintf", strerror(errno));
-				fclose(changefile_fptr);
+				(void) fclose(changefile_fptr);
 				break;
 			}
 			if (lstat(changedpath, &sb) == 0)
-				fprintf(list_fptr, "%s\n", linebuf);
+				fprintf(list_fptr, "%s%c", linebuf_ptr, line_delimiter);
 			free(changedpath);
 		}
 
-		fclose(changefile_fptr);
+		(void) fclose(changefile_fptr);
+
+		if (NULL != linebuf_ptr)
+			free(linebuf_ptr);
 
 		remove(path);
+		free(path);
 	}
 
 	if (NULL != tree_root)
@@ -961,7 +1073,9 @@ static void collate_transfer_list(struct sync_set_s *cf)
 	}
 	free(namelist);
 
-	fclose(list_fptr);
+	if (0 != fclose(list_fptr)) {
+		error("%s: %s", cf->transfer_list, strerror(errno));
+	}
 }
 
 
@@ -976,14 +1090,17 @@ static void collate_transfer_list(struct sync_set_s *cf)
  *
  * The status file is updated before an action is performed.
  */
-static int sync_partial(struct sync_set_s *cf, struct sync_status_s *st)
+static int sync_partial(struct sync_set_s *cf, struct sync_status_s *st, bool rsync_null_delimiter)
 {
 	struct stat sb;
 	int lockfd = -1;
 	int rc = 0;
 	FILE *list_fptr;
+	char line_delimiter;
 
-	collate_transfer_list(cf);
+	line_delimiter = rsync_null_delimiter ? '\0' : '\n';
+
+	collate_transfer_list(cf, rsync_null_delimiter);
 
 	if ((stat(cf->transfer_list, &sb) != 0) || (0 == sb.st_size)) {
 		/*
@@ -1022,32 +1139,47 @@ static int sync_partial(struct sync_set_s *cf, struct sync_status_s *st)
 	 */
 	list_fptr = fopen(cf->transfer_list, "r");
 	if (NULL != list_fptr) {
-		char linebuf[4096];
-		int lineno = 0;
-		while ((!feof(list_fptr))
-		       && (NULL != fgets(linebuf, sizeof(linebuf) - 1, list_fptr))) {
-			char *nlptr;
+		char *linebuf_ptr = NULL;
+		size_t linebuf_size = 0;
+		unsigned int line_number = 0;
 
-			lineno++;
-			if (lineno > 100) {
+		while (0 == feof(list_fptr)) {
+			ssize_t line_length = 0;
+
+			/* Same logic as in collate_transfer_list(). */
+			errno = 0;
+			line_length = getdelim(&linebuf_ptr, &linebuf_size, line_delimiter, list_fptr);
+			if ((line_length < 0) || (NULL == linebuf_ptr)) {
+				if (0 != errno)
+					error("%s:%u: %s", cf->transfer_list, line_number, strerror(errno));
+				break;
+			}
+			line_number++;
+			if (line_length < 1)
+				continue;
+			line_length--;
+			linebuf_ptr[line_length] = '\0';
+
+			if (line_number > 100) {
 				log_message(cf->log_file, "[%s]   %s", cf->name, "...");
 				break;
 			}
 
-			nlptr = strrchr(linebuf, '\n');
-			if (NULL != nlptr)
-				nlptr[0] = '\0';
-
-			log_message(cf->log_file, "[%s]   %s", cf->name, linebuf);
+			log_message(cf->log_file, "[%s]   %s", cf->name, linebuf_ptr);
 		}
-		fclose(list_fptr);
+
+		if (NULL != linebuf_ptr)
+			free(linebuf_ptr);
+
+		(void) fclose(list_fptr);
 	}
 
 	rc = run_rsync(cf->log_file, cf->name, cf->source, cf->destination,
 		       st->excludes_file,
 		       NULL ==
 		       cf->partial_rsync_opts ? "--delete -dlptgoDH" :
-		       cf->partial_rsync_opts, cf->transfer_list, cf->ignore_vanished_files, st->rsync_error_file);
+		       cf->partial_rsync_opts, cf->transfer_list, cf->ignore_vanished_files, rsync_null_delimiter,
+		       st->rsync_error_file, cf->partial_timeout);
 
 	log_message(cf->log_file, "[%s] %s: %s: %s", cf->name,
 		    _("partial sync"), _("sync ended"), rc == 0 ? _("OK") : _("FAILED"));
@@ -1065,7 +1197,7 @@ static int sync_partial(struct sync_set_s *cf, struct sync_status_s *st)
 		update_timestamp_file(cf, cf->partial_marker);
 		st->last_partial_sync = time(NULL);
 		st->partial_sync_failures = 0;
-		st->last_partial_sync_status = "OK";
+		st->last_partial_sync_status = STRING_STATUS_OK;
 	}
 
 	return rc;
@@ -1075,7 +1207,7 @@ static int sync_partial(struct sync_set_s *cf, struct sync_status_s *st)
 /*
  * Log the given message to the given file, with a timestamp.
  */
-static void log_message(const char *file, const char *format, ...)
+static void log_message( /*@null@ */ const char *file, const char *format, ...)
 {
 	FILE *fptr;
 	va_list ap;
@@ -1127,7 +1259,10 @@ static void log_message(const char *file, const char *format, ...)
 		if (0 == lockf(fileno(fptr), F_ULOCK, 0))
 			locked = false;
 	}
-	fclose(fptr);
+
+	if (0 != fclose(fptr)) {
+		debug("(log) %s: %s", file, strerror(errno));
+	}
 }
 
 
